# $Id$
# Copyright 1999 Philip Gwyn
# This code distributed under the same Licence as Perl itself.

package XML::Storable;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $study %seen %selfref);
use Data::Dumper;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	xml_freeze
	xml_thaw
    expat_thaw
);
@EXPORT_OK=qw(
    xml_freeze
	xml_thaw
    xml_study
);
$VERSION = '0.03';
$study=1;

#######################################################################
sub freeze
{
    my $ret='';
    local(%seen);
    local %selfref if $study;
    xml_study(@_) if $study;

    foreach my $e (@_)
    {
        $ret.=_freeze($e);
    }
    return $ret;
}
*xml_freeze=\&freeze;

#######################################################################
sub xml_study
{
    foreach my $e (@_)
    {
        next unless ref $e;
        $selfref{$e}++;
        next if $selfref{$e}>1;
        xml_study(@$e)          if $e=~/ARRAY/;
        xml_study(values %$e)   if $e=~/HASH/;
    }
}

#######################################################################
sub _freeze
{
    my $e=shift;
    return "<UNDEF></UNDEF>" if not defined $e;
    my $type=ref $e;
    my $ret;

    ## plain scalar
    return '<S>'._xml_safe($e).'</S>' if not $type; 

    $ret='<'.$type;                     # commence building start tag

    
    my $class='';
    my $id;

    ## Blessed objects need to be jiggery pokered...
    if($e=~m(^$type=(HASH|ARRAY|SCALAR)\x280x([0-9A-Fa-f]+)\x29$))
    {
        # return $e->xml_freeze() if $e->can('xml_freeze');
        $class=' Class="'.$type.'"';
        $ret.=$class;
        $type=$1;
        $id=$2;
    }
          ## Get the reference's unique key
    elsif(not defined %selfref or          # no study OR
                ($selfref{$e}||0)>1)             # seen >1
    {
        $id=$1 if $e=~m(^$type\x280x([0-9A-Fa-f]+)\x29$);
        die "Why isn't $e a $type!" unless $id;
    }
    $ret.=' ID="'.$id.'"' if $id;
    $ret.='>';

    ## Now deal with the real data
    if($type eq 'HASH')
    {
        if($id)
        {
            return '<HASH Ref="'.$id.'"></HASH>' if $seen{$id};
            $seen{$id}=1;
        }
        while(my($k, $v)=each %$e)
        {
            $ret.='<KEY>'._xml_safe($k).'</KEY><VALUE>'._freeze($v).'</VALUE>';
        }
        $ret.='</HASH>';

    } elsif($type eq 'ARRAY')
    {
        if($id)
        {
            return '<ARRAY Ref="'.$id.'"></ARRAY>' if $seen{$id};
            $seen{$id}=1;
        }
        local $_;
        foreach (@$e)
        {   $ret.=_freeze($_); }

        $ret.='</ARRAY>';    

    } elsif($type eq 'SCALAR')
    {
        if($id)
        {
            return '<SCALAR Ref="'.$id.'"></SCALAR>' if $seen{$id};
            $seen{$id}=1;
        }
        $ret.=_xml_safe($e).'</SCALAR>';  

    } else
    {
        die "Unknown type: $type\n";
    }
    return $ret;
}

#######################################################################
## Stragegy used : 
##    We build up a value in @accume.  When a new tag is started, pertinent
## information (tag, attributes, position in @accume) are kept in @stack.
## When a tag is closed, the value is splice()ed off @accume to form the $body
## (always a ref, except for <S> tags).
##
## ID= tags are kept in %seen and a Ref= is then pulled from there
sub perl_thaw
{
    my($text)=@_;
    study $text;        # Gives us 16% win
    my(%seen);

    my $hash=0;
    my($type, $id, @accume, $att, $body, @stack, $end, $q, $pos);
    while($text=~m(<(/)?                            # possible end
                    (ARRAY|HASH|S|SCALAR|UNDEF|KEY|VALUE)  # tag
                    ([^>]*?)>                       # attributes
                    ([^<]*))gx)                    # and text-body
    {
        $end=$1; $type=$2; $att=$3; $body=$4;        
        study $att;
        if($end)            # END of a tag
        {
            ($q, $att, $body, $pos, $hash, $id)=@{pop @stack};
            
            warn "<$q> ended with </$type>" if($q ne $type);
            $type=$q;

            if($att=~m(\bRef="([0-9a-fA-F]+)"))
            {
                push @accume, $seen{$1};
                next;
            }
            
            if($type eq 'KEY')              ### hash key
            {
                if($hash==0)                # are we in a hash at least
                {   warn "<KEY> tag not in <HASH>"; } 
                elsif($hash==2)             # can't have <KEY><KEY>...
                {   warn "<KEY> found, expecting <VALUE>"; }
                $hash=2;
                $type='S';                  # treat the KEY as a scalar
            } 
            if($type eq 'VALUE')            ### hash value
            {
                $body=pop @accume;          
                if($pos != @accume)         # *must* go back to where we were
                {
                    warn "<VALUE> contained more then one element\n";
                    splice @accume, $pos;
                }

                if($hash==0)                # are we in a hash...
                {   warn "<KEY> tag not in <HASH>\n";
                } elsif($hash==1)           # have we seen the key at least?
                {   warn "<VALUE> found, expecting <KEY>\n";
                    push @accume, 'KEY';    # bogus key for this value
                }
                $hash=1;                    # now looking for KEY
                
            } elsif($type eq 'S')           ### simple scalar
            {
                $body=_xml_unsafe($body);

            } elsif($type eq 'HASH')        ### hash ref
            {
                if($hash==2)
                {
                    warn "Missing <VALUE>\n";
                    $#accume++;
                }
                $body={splice @accume, $pos};
                if($id and $seen{$id})
                {
                    %{$seen{$id}}=%$body;   # copy the body
                    $body=$seen{$id};
                }
            } elsif($type eq 'ARRAY')       ### array ref
            {
                $body=[splice @accume, $pos];
                if($id and $seen{$id})
                {
                    @{$seen{$id}}=@$body;   # copy the body
                    $body=$seen{$id};
                }
    
            } elsif($type eq 'SCALAR')      ### scalar ref
            {
                $body=_xml_unsafe($body);
                $body=\$body;
                if($id and $seen{$id})
                {
                    ${$seen{$id}}=$$body;   # copy the body
                    $body=$seen{$id};
                }

            } elsif($type eq 'UNDEF')       ### undef
            {
                undef $body;
            } else
            {
                warn "What the hell is a $type ($body)";
            }
                                            ## bless it if need be
            bless $body, $2 if $att=~m(\bClass="([^"]+)"); 
            if($id)                         ### remember that we've seen it...
            {
                unless($seen{$id})          # already refrenced!
                {
                    $seen{$id}=$body ;      # save it
                }
            }
            push @accume, $body;            ### keep it on the output stack
        } else
        {
            $att=~m(\bID="([0-9a-fA-F]+)"); # wish i didn't have to do this
            $id=$1;                         # for every tag
            if($id and not $seen{$id})
            {
                $seen{$id}=($type eq 'HASH' ? {} : 
                            $type eq 'ARRAY' ? [] : \''
                           );
            } else
            {
                # die qq(<$type ID="$id"> seen more then once!);
            }
            push @stack, [$type, $att, $body, scalar(@accume), $hash, $id];
            $hash=($type eq 'HASH');
        }
    }
    return @accume if wantarray;
    return $accume[0] if 1==scalar @accume;
    return \@accume;
}

#######################################################################
sub expat_thaw
{
    my($text)=@_;
    return unless $text;
    
    my $heap={stack=>[], accume=>[], seen=>{}, hash=>0};
    my $p=XML::Parser->new(ErrorContext=>2,
                           Handlers=>{
                                Start=>sub{_ex_start($heap, @_)},
                                End=>sub{_ex_end($heap, @_)},
                                Char=>sub{_ex_char($heap, @_)},
                                Default=>sub{_ex_default($heap, @_)},
                            },
                            ProtocolEncoding=>'UTF-8',
                            NoExpand=>1,
                          );
    $p->parsestring('<?xml version="1.0"?>'.$text);
    return @{$heap->{accume}} if wantarray;
    return $heap->{accume}[0] if 1==scalar @{$heap->{accume}};
    return $heap->{accume};
}

#################################
sub _ex_start
{
    my($heap, $xp, $element, %att)=@_;
    die "Unknown element $element" 
                unless $element =~ /^ARRAY|HASH|S|SCALAR|UNDEF|KEY|VALUE$/;
    if($att{ID} and not $heap->{seen}{$att{ID}})
    {
        $heap->{seen}{$att{ID}}=($element eq 'HASH' ? {} : 
                                     $element eq 'ARRAY' ? [] : \''
                                    );
    } else
    {
        # die qq(<$type ID="$id"> seen more then once!);
    }
    push @{$heap->{stack}}, 
                    [$element, scalar(@{$heap->{accume}}), $heap->{hash},
                      $att{ID}, $att{Class}, ($att{'ref'}||$att{Ref}), ''];
    $heap->{current}=\$heap->{stack}[-1][-1];
    $heap->{hash}=($element eq 'HASH')?1:0;
}

#################################
sub _ex_end
{
    my($heap,  $xp, $element)=@_;
    my($type, $pos, $hash, $id, $class, $ref, $body)=
            @{pop @{$heap->{stack}}};
    $heap->{current}=\$heap->{stack}[-1][-1] if @{$heap->{stack}};
    $heap->{hash}=$hash;


    if($ref)
    {
        if($pos != @{$heap->{accume}})      # *must* go back to where we were
        {
            warn qq(<$element Ref="$ref"> must be empty!);
            splice @{$heap->{accume}}, $pos;
        }
        push @{$heap->{accume}}, $heap->{seen}{$ref};
        return;
    }

    if($type eq 'KEY')              ### hash key
    {
        if($hash==0)                # are we in a hash at least
        {   warn "<KEY> tag not in <HASH>"; } 
        elsif($hash==2)             # can't have <KEY><KEY>...
        {   warn "<KEY> found, expecting <VALUE>"; }
        $heap->{hash}=2;
        $body=_xml_unsafe($body);
    } elsif($type eq 'VALUE')            ### hash value
    {
        $body=pop @{$heap->{accume}};
        if($pos != @{$heap->{accume}})      # *must* go back to where we were
        {
            warn "<VALUE> contained more then one element\n";
            splice @{$heap->{accume}}, $pos;
        }

        if($hash==0)                        # are we in a hash...
        {   warn "<KEY> tag not in <HASH>\n";
        } elsif($hash==1)                   # have we seen the key at least?
        {   warn "<VALUE> found, expecting <KEY>\n";
            push @{$heap->{accume}}, 'KEY'; # bogus key for this value
        }
        $heap->{hash}=1;                    # now looking for KEY
                
    } elsif($type eq 'S')           ### simple scalar
    {
        $body=_xml_unsafe($body);

    } elsif($type eq 'HASH')        ### hash ref
    {
        if($hash==2)
        {
            warn "Missing <VALUE>\n";
            $#{$heap->{accume}}++;
        }
        $body={splice @{$heap->{accume}}, $pos};
        if($id and $heap->{seen}{$id})
        {
            %{$heap->{seen}{$id}}=%$body;   # copy the body
            $body=$heap->{seen}{$id};
        }
    } elsif($type eq 'ARRAY')       ### array ref
    {
        $body=[splice @{$heap->{accume}}, $pos];
        if($id and $heap->{seen}{$id})
        {
            @{$heap->{seen}{$id}}=@$body;   # copy the body
            $body=$heap->{seen}{$id};
        }
    
    } elsif($type eq 'SCALAR')              ### scalar ref
    {
        $body=_xml_unsafe($body);
        $body=\$body;
        if($id and $heap->{seen}{$id})
        {
            ${$heap->{seen}{$id}}=$$body;   # copy the body
            $body=$heap->{seen}{$id};
        }

    } elsif($type eq 'UNDEF')               ### undef
    {
        undef $body;
        $class='';                          # don't bless it
    } else
    {
        warn "What the hell is a ".Dumper($type);
    }

    bless $body, $class if $class;          ## bless it if need be
    if($id)                                 ### remember that we've seen it...
    {
        unless($heap->{seen}{$id})          # already refrenced!
        {
            $heap->{seen}{$id}=$body ;      # save it
        }
    }
    push @{$heap->{accume}}, $body;         ### keep it on the output stack
}


#################################
sub _ex_char
{
    my($heap,  $xp, $string)=@_;
    return unless $string;
    ${$heap->{current}}.=$string;
}

#################################
sub _ex_default
{
    my($heap,  $xp, $string)=@_;
#    warn "What to do with ".Dumper($string);
}

#######################################################################
BEGIN
{
    eval { require XML::Parser; };
    if(($@||'')=~/Can't locate/)
    {
        # warn "Using Perl implementation";
        *thaw=\&perl_thaw;
    } elsif($@)
    {
        die $@;
    } else
    {
        # warn "Using XML::Parser";
        *thaw=\&expat_thaw;
    }
    *xml_thaw=\&thaw;
}

#######################################################################
sub _xml_safe
{
    my($val)=@_;
    $val=~s(&)(&#38;)g;     # MUST BE FIRST
    $val=~s(/)(&#47;)g;
    $val=~s(<)(&#60;)g;
    $val=~s(>)(&#62;)g; 
    $val=~s(([\x00-\x1f]))('&#'.ord($1).';')ge; 
    return $val;
}

#######################################################################
sub _xml_unsafe
{
    my($val)=@_;
    $val=~s(&#(\d+);)(chr($1))ge; 
    $val=~s(&#38;)(&)g;     # MUST BE LAST
    return $val;
}


1;
__END__
#######################################################################
$Log$


=head1 NAME

XML::Storable - converting Perl structures to XML and back

=head1 SYNOPSIS

    use XML::Storable;
    my $stored=xml_freeze(\%table, \@chair);
    ... send it over a socket or something ...
    my ($table, $chair)=xml_thaw($stored);

=head1 DESCRIPTION

Convert data to/from an XML representation for throwing over the net or 
storeing to almost human readable files.

=head2 xml_freeze($data, ...)

Converts the input values into XML::Storable format.  No effort is made
at pretty-printing the output, mostly for speed reasons but also because
the output is only supposed to be parsed by other programs.

=over 4

=item xml_freeze('hello'); produces:

    <S>hello</S>

=item xml_freeze(\'hello'); produces:

    <SCALAR ID=34134230>hello</SCALAR>

=item xml_freeze([0..9]); produces:

    <ARRAY ID=023c4af3><S>0</S><S>1</S><S>2</S><S>3</S><S>4</S>
    <S>5</S><S>6</S><S>7</S><S>8</S><S>9</S></ARRAY>

only without the newline.

=item xml_freeze({foo=>'bar', bif=>'boff'}); produces

    <HASH ID=1234568><KEY>foo</KEY><VALUE><S>bar</S></VALUE>
    <KEY>bif</KEY><VALUE><S>boff</VALUE></HASH>

only without the newline.

=back

xml_freeze can also handle self-referencial structures.

    my $h={
        name=>"King George",
        date=>"1999\n&\n<FOO>",
        spouses=>[
            qw(Henrietta Julietta Paula Pili Pollo)
        ],

    };
    $h->{myself}=$h;
    print xml_freeze($h), "\n";

produces:
    
    <HASH ID=80dde9c><KEY>myself</KEY>
    <VALUE><HASH Ref="80dde9c"></></VALUE>
    <KEY>name</KEY><VALUE><S>King George</S></VALUE>
    <KEY>spouses</KEY><VALUE><ARRAY ID=80dde48>
    <S>Henrietta</S><S>Julietta</S><S>Paula</S>
    <S>Pili</S><S>Pollo</S></ARRAY></VALUE>
    <KEY>date</KEY><VALUE>
    <S>1999&#10;&#38;&#10;&#60;FOO&#62;</S></VALUE>
    </HASH>

only without the newlines.


=head2 xml_thaw($xml)

Turns the XML into a perl structure.  A special  XML parser is used. It has
been hand coded for our specific DTD and cuts some corners very close.  If
you are feeding it data from another source, please read the L<BUGS> section
below.

=head2 xml_study($xml)

Recurses through the structure to see if it is self-referencial.  Elements
that need references will have ID= attributes.  This slows freeze down a bit
but makes thaw run faster.  Almost.

xml_study() uses a global variable that is not reset.

=head2 $XML::Storable::study

If set (default), xml_study() is called at each invocation of 
xml_freeze().

If not set, xml_study() is not automaticaly called.  You may call it on
your data yourself.  If you do not, and ID= tag will be generated for all 
references, even if they aren't needed.

=head1 BUGS

Will fail on coderefs and tied strutures.

The XML parser is hand coded and not very good at error handling.  XML is
expected to be produced by XML::Storable::freeze.  For instance, ID= and
Ref= to have "" around the value.  And a <TAG> with a Ref=
attribute should not be closed with a </TAG>.  Please use <TAG Ref=foo></> 
(this is the only case where </> may be used).

The XML parser can't handle forward references (ie <HASH Ref=foo> but 
<HASH ID=foo> is later on in the string.

It is currently very slow, 10x slower then Storable and about 2x slower then
FreezeThaw.  ./benchmark for detail.  But then Storable uses an XS to get
it's speed.

An attempt has been made to preserve blessed objects.  However, 
XML::Storable cheats and looks behind the scene to see if you use a hashref, 
arrayref or scalarref.  Eventually, it should try to use 
xml_(freeze,thaw,study) methods if they are available to that object;

=head1 AUTHOR

Philip Gwyn, fil@pied.nu

=head1 SEE ALSO

L<FreezeThaw>, L<Storable>, L<XML::Parser>

=cut
